<?php

declare(strict_types=1);

namespace App\Arbitration\Inner\Stuff;

use App\Arbitration\Inner\ArbitrationChain;
use App\Exchanges\ExchangeInterface;
use App\Exchanges\MarketInterface;
use App\Utils\Money\CurrencyPair;

class ArbitrationMarketChainsFinder
{

    /**
     * @param ExchangeInterface $exchange
     * @param int $maxChainDepth
     * @return ArbitrationChain[]
     */
    public function findChainsInExchange(ExchangeInterface $exchange, int $maxChainDepth = 3): array
    {
        /**
         * @var MarketInterface[] $markets
         */
        $markets = $exchange->getMarkets();

        [$rows, $map] = $this->buildMap($markets);

        $cycles = [];
        foreach (array_keys($rows) as $row) {
            $rowCycles = [];
            $this->findCycle($map, [$row], $rowCycles, $maxChainDepth);
            $cycles[] = $this->clearDuplicates($rowCycles);
        }

        $cycles = array_merge(...$cycles);
        $cycles = $this->clearDuplicates($cycles);

        $result = [];
        foreach ($cycles as $cycle) {
            $result[] = $this->cycleToChain(
                $cycle,
                $rows,
                $markets
            );
        }

        return $result;
    }

    protected function cycleToChain(array $cycle, array $rows, array $markets): ArbitrationChain
    {
        $cycleMarkets = [];
        for($i = 0; $i < \count($cycle); $i++) {
            $base = $rows[$cycle[$i]];
            if ($i + 1 >= \count($cycle)) {
                $quote = $rows[$cycle[0]];
            } else {
                $quote = $rows[$cycle[$i + 1]];
            }

            foreach ($markets as $market) {
                /**
                 * @var CurrencyPair $pair
                 */
                $pair = $market->getPair();

                if (
                    (
                        $pair->getBase()->isEqual($base)
                        && $pair->getQuote()->isEqual($quote)
                    )
                    || (
                        $pair->getBase()->isEqual($quote)
                        && $pair->getQuote()->isEqual($base)
                    )
                ) {
                    $cycleMarkets[] = $market;
                }
            }
        }

        return new ArbitrationChain(...$cycleMarkets);
    }

    /**
     * @param MarketInterface[] $markets
     * @return array
     */
    protected function buildMap(array $markets): array
    {
        $map = [];
        $rows = [];
        foreach ($markets as $market) {
            $pair = $market->getPair();

            $base = $pair->getBase();
            if (!in_array((string)$base, $rows)) {
                $rows[] = $base;
            }
            $baseId = array_search((string)$base, $rows);

            $quote = $pair->getQuote();
            if (!in_array((string)$quote, $rows)) {
                $rows[] = $quote;
            }
            $quoteId = array_search((string)$quote, $rows);

            $map[$quoteId][$baseId] = 1;
            $map[$baseId][$quoteId] = 1;
        }

        return [$rows, $map];
    }

    protected function clearDuplicates(array $cycles): array
    {
        $orderedCycles = [];
        foreach ($cycles as $i => $cycle) {
            $orderedCycle = $cycle;
            sort($orderedCycle);
            $orderedCycle = implode(',', $orderedCycle);
            if (in_array($orderedCycle, $orderedCycles)) {
                unset($cycles[$i]);
                continue;
            }
            $orderedCycles[] = $orderedCycle;
        }

        return $cycles;
    }

    protected function findCycle(array $map, array $path, array &$result, int $maxDepth)
    {
        if (\count($path) > $maxDepth) {
            return;
        }
        $key = $path[\count($path) - 1];
        $row = $map[$key];


        unset($row[$key]);
        foreach ($row as $next => $v) {
            if (array_search($next, $path)) {
                continue;
            }
            $nextPath = $path;
            $nextPath[] = $next;

            if (\count($path) > 2 && $path[0] === $next) {
                $result[] = $path;
                continue;
            }

            $this->findCycle($map, $nextPath, $result, $maxDepth);
        }
    }
}